# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import gobject

class FilteredTree():
    """ FilteredTree is the most important and also the most buggy part of
    LibLarch.

    FilteredTree transforms general changes in tree like creating/removing 
    relationships between nodes and adding/updating/removing nodes into a serie
    of simple steps which can be for instance by GTK Widget.

    FilteredTree allows filtering - hiding certain nodes defined by a predicate.

    The reason of most bugs is that FilteredTree is request to update a node.
    FilteredTree must update its ancestors and also decestors. You cann't do that
    by a simple recursion.
    """

    def __init__(self, tree, filtersbank, name=None, refresh=True):
        """ Construct a layer where filters could by applied

        @param tree: Original tree to filter.
        @param filtersbank: Filter bank which stores filters
        @param refresh: Requests all nodes in the beginning? Additional
            filters can be added and refresh can be done later

        _flat defines whether only nodes without children can be shown. For example WorkView filter.
        """

        self.cllbcks = {}

        # Cache
        self.nodes = {}
        # Set root_id by the name of FilteredTree
        if name is None:
            self.root_id = "anonymous_root"
        else:
            self.root_id = "root_%s" % name

        self.nodes[self.root_id] = {'parents': [], 'children': []}
        self.cache_paths = {}
        self.filter_cache = {}

        # Connect to signals from MainTree
        self.tree = tree
        self.tree.register_callback("node-added", self.__external_modify)
        self.tree.register_callback("node-modified", self.__external_modify)
        self.tree.register_callback("node-deleted", self.__external_modify)

        # Filters
        self.__flat = False
        self.applied_filters = []
        self.fbank = filtersbank
        
        if refresh:
            self.refilter()
            

    def set_callback(self, event, func,node_id=None, param=None):
        """ Register a callback for an event.

        It is possible to have just one callback for event.
        @param event: one of added, modified, deleted, reordered
        @param func: callback function
        """
        if event == 'runonce':
            if not node_id:
                raise Exception('runonce callback should come with a node_id')
            if self.is_displayed(node_id):
                #it is essential to idle_add to avoid hard recursion
                gobject.idle_add(func,param)
            else:
                if not self.cllbcks.has_key(node_id):
                    self.cllbcks[node_id] = []
                self.cllbcks[node_id].append([func,node_id,param])
        else:
            self.cllbcks[event] = [func,node_id,param]
        
    def callback(self, event, node_id, path, neworder=None):
        """ Run a callback.

        To call callback, the object must be initialized and function exists.

        @param event: one of added, modified, deleted, reordered, runonce
        @param node_id: node_id parameter for callback function
        @param path: path parameter for callback function
        @param neworder: neworder parameter for reorder callback function
        
        The runonce event is actually only run once, when a given task appears.
        """

        
        if event == 'added':
            for func,nid,param in self.cllbcks.get(node_id,[]):
                if nid and self.is_displayed(nid):
                    func(param)
                    if self.cllbcks.has_key(node_id):
                        self.cllbcks.pop(node_id)
                else:
                    raise Exception('%s is not displayed but %s was added' %(nid,node_id))
        func,nid,param = self.cllbcks.get(event, (None,None,None))
        if func:
            if neworder:
                func(node_id,path,neworder)
            else:
                func(node_id,path)
                    

#### EXTERNAL MODIFICATION ####################################################
    def __external_modify(self, node_id):
        return self.__update_node(node_id,direction="both")
        
    def __update_node(self, node_id, direction):
        '''update the node node_id and propagate the 
        change in direction (up|down|both) '''

        if node_id == self.root_id:
            return None

        current_display = self.is_displayed(node_id)
        new_display = self.__is_displayed(node_id)

        for fcname in self.filter_cache:
            if node_id in self.filter_cache[fcname]['nodes']:
                self.filter_cache[fcname]['nodes'].remove(node_id)
                self.filter_cache[fcname]['count'] = len(self.filter_cache[fcname]['nodes'])

        completely_updated = True

        if not current_display and not new_display:
            # If a task is not displayed and should not be displayed, we
            # should still check its parent because he might not be aware
            # that he has a child
            if self.tree.has_node(node_id):
                node = self.tree.get_node(node_id)
                for parent in node.get_parents():
                    self.__update_node(parent,"up")
            return completely_updated
        elif not current_display and new_display:
            action = 'added'
        elif current_display and not new_display:
            action = 'deleted'
        else:
            action = 'modified'

        # Create node info for new node
        if action == 'added':
            self.nodes[node_id] = {'parents':[], 'children':[]}

        # Make sure parents are okay if we adding or updating
        if action == 'added' or action == 'modified':
            current_parents = self.nodes[node_id]['parents']
            new_parents = self.__node_parents(node_id)

            # When using flat filter or a recursive filter, FilteredTree
            # might not recognize a parent correctly, make sure to check them
            if action == 'added':
                node = self.tree.get_node(node_id)
                for parent_id in node.get_parents():
                    if parent_id not in new_parents and parent_id not in current_parents:
                        self.__update_node(parent_id, direction="up")

            # Refresh list of parents after doing checkup once again
            current_parents = self.nodes[node_id]['parents']
            new_parents = self.__node_parents(node_id)
                    
            self.nodes[node_id]['parents'] = [parent_id for parent_id in new_parents
                if parent_id in self.nodes]

            remove_from = list(set(current_parents) - set(new_parents))
            add_to = list(set(new_parents) - set(current_parents))
            stay = list(set(new_parents) - set(add_to))

            #If we are updating a node at the root, we should take care
            #of the root too
            if direction == "down" and self.root_id in add_to:
                direction = "both"

            for parent_id in remove_from:
                self.send_remove_tree(node_id, parent_id)
                self.nodes[parent_id]['children'].remove(node_id)
                if direction == "both" or direction == "up":
                    self.__update_node(parent_id,direction="up")
            #there might be some optimization here
            for parent_id in add_to:
                if parent_id in self.nodes:
                    self.nodes[parent_id]['children'].append(node_id)
                    self.send_add_tree(node_id, parent_id)
                    if direction == "both" or direction == "up":
                        self.__update_node(parent_id,direction="up")
                else:
                    completely_updated = False
                    raise Exception("We have a parent not in the ViewTree")
            #We update all the other parents
            if direction == "both" or direction == "up":
                for parent_id in stay:
                    self.__update_node(parent_id,direction="up")
            #We update the node itself     
            #Why should we call the callback only for modify?
            if action == 'modified':
                for path in self.get_paths_for_node(node_id):
                    self.callback(action, node_id, path) 
            
            #We update the children
            current_children = self.nodes[node_id]['children']
            new_children = self.__node_children(node_id)
            if direction == "both" or direction == "down":
                for cid in new_children:
                    if cid not in current_children:
                        self.__update_node(cid,direction="down")

        elif action == 'deleted':
            paths = self.get_paths_for_node(node_id)
            children = list(reversed(self.nodes[node_id]['children']))
            for child_id in children:
                self.send_remove_tree(child_id, node_id)
                self.nodes[child_id]['parents'].remove(node_id)
                self.__update_node(child_id,direction="down")

            for path in paths:
                self.callback(action, node_id, path)

            # Remove node from cache
            for parent_id in self.nodes[node_id]['parents']:
                self.nodes[parent_id]['children'].remove(node_id)
                self.__update_node(parent_id,direction="up")

            self.nodes.pop(node_id)

            # We update parents who are not displayed
            # If the node is only hidden and still exists in the tree
            if self.tree.has_node(node_id):
                node = self.tree.get_node(node_id)
                for parent in node.get_parents():
                    if parent not in self.nodes:
                        self.__update_node(parent, direction="up")
                
        return completely_updated

    def send_add_tree(self, node_id, parent_id):
        paths = self.get_paths_for_node(parent_id)
        queue = [(node_id, (node_id, ))]

        while queue != []:
            node_id, relative_path = queue.pop(0)

            for start_path in paths:
                path = start_path + relative_path
                self.callback('added', node_id, path)

            for child_id in self.nodes[node_id]['children']:
                queue.append((child_id, relative_path + (child_id,)))

    def send_remove_tree(self, node_id, parent_id):
        paths = self.get_paths_for_node(parent_id)
        stack = [(node_id, (node_id, ), True)]

        while stack != []:
            node_id, relative_path, first_time = stack.pop()

            if first_time:
                stack.append((node_id, relative_path, False))
                for child_id in self.nodes[node_id]['children']:
                    stack.append((child_id, relative_path + (child_id,), True))

            else:
                for start_path in paths:
                    path = start_path + relative_path
                    self.callback('deleted', node_id, path)

    def test_validity(self):
        for node_id in self.nodes:
            for parent_id in self.nodes[node_id]['parents']:
                assert node_id in self.nodes[parent_id]['children'], "Node '%s' is not in children of '%s'" % (node_id, parent_id)

            if self.nodes[node_id]['parents'] == []:
                assert node_id == self.root_id, "Node '%s' does not have parents" % (node_id)

            for parent_id in self.nodes[node_id]['children']:
                assert node_id in self.nodes[parent_id]['parents'], "Node '%s' is not in parents of '%s'" % (node_id, parent_id)


#### OTHER ####################################################################
    def refilter(self):
        # Find out it there is at least one flat filter
        self.filter_cache = {}
        self.__flat = False
        for filter_name in self.applied_filters:
            filt = self.fbank.get_filter(filter_name)
            if filt and not self.__flat:
                self.__flat = filt.is_flat()

        # Clean the tree
        for node_id in reversed(self.nodes[self.root_id]['children']):
            self.send_remove_tree(node_id, self.root_id)

        self.nodes = {}
        self.nodes[self.root_id] = {'parents': [], 'children': []}

        # Build tree again
        root_node = self.tree.get_root()
        queue = root_node.get_children()

        while queue != []:
            node_id = queue.pop(0)
            # FIXME: decide which is the best direction
            self.__update_node(node_id, direction="both")

            node = self.tree.get_node(node_id)
            for child_id in node.get_children():
                queue.append(child_id)

    def __is_displayed(self, node_id):
        """ Should be node displayed regardless of its current status? """
        if node_id and self.tree.has_node(node_id):
            for filter_name in self.applied_filters:
                filt = self.fbank.get_filter(filter_name)
                if filt:
                    can_be_displayed = filt.is_displayed(node_id)
                    if not can_be_displayed:
                        return False
                else:
                    return False
            return True
        else:
            return False

    def is_displayed(self, node_id):
        """ Is the node displayed at the moment? """

        return node_id in self.nodes

    def __node_children(self, node_id):
        if node_id == self.root_id:
            raise Exception("Requesting children for root node")

        if not self.__flat:
            if self.tree.has_node(node_id):
                node = self.tree.get_node(node_id)
            else:
                node = None
        else:
            node = None

        if not node:
            return []

        toreturn = []
        for child_id in node.get_children():
            if self.__is_displayed(child_id):
                toreturn.append(child_id)

        return toreturn

    def __node_parents(self, node_id):
        """ Returns parents of the given node. If node has no parent or 
        no displyed parent, return the virtual root.
        """
        if node_id == self.root_id:
            raise ValueError("Requested a parent of the root node")

        parents_nodes = []
        #we return only parents that are not root and displayed
        if not self.__flat and self.tree.has_node(node_id):
            node = self.tree.get_node(node_id)
            for parent_id in node.get_parents():
                if parent_id in self.nodes and self.__is_displayed(parent_id):
                    parents_nodes.append(parent_id)

        # Add to root if it is an orphan
        if parents_nodes == []:
            parents_nodes = [self.root_id]

        return parents_nodes
        
    #This is a crude hack which is more performant that other methods
    def is_path_valid(self,p):
#        print "is %s valid?" %str(p)
        valid = True
        i = 0
        if len(p) == 1:
            valid = False
        else:
            while valid and i < len(p) - 1:
                child = p[i+1]
                par = p[i]
                if self.nodes.has_key(par):
                    valid = (child in self.nodes[par]['children'])
                else:
                    valid = False
                i += 1
        return valid

    def get_paths_for_node(self, node_id):
#        cached = self.cache_paths.get(node_id,None)
        #The cache improves performance a lot for "stairs"
        #FIXME : the cache cannot detect if a new path has been added
#        validcache = False
#        if cached:
#            validcache = True
#            for p in cached:
#                validcache = validcache and self.is_path_valid(p)
#            if validcache:
##                print "the valid cache is : %s" %str(cached)
#                return cached

        
        if node_id == self.root_id or not self.is_displayed(node_id):
            return [()]
        else:
            toreturn = []
            for parent_id in self.nodes[node_id]['parents']:
                if parent_id not in self.nodes:
                    raise Exception("Parent %s does not exists" % parent_id)
                if node_id not in self.nodes[parent_id]['children']:
                    # Dump also state of FilteredTree => useful for debugging
                    s = "\nCurrent tree:\n"
                    for key in self.nodes:
                        s += key + "\n"
                        s += "\t parents" + str(self.nodes[key]['parents']) + "\n"
                        s += "\t children" + str(self.nodes[key]['children']) + "\n"
                    raise Exception("%s is not children of %s\n%s" % (node_id, parent_id, s))

                for parent_path in self.get_paths_for_node(parent_id):
                    mypath = parent_path + (node_id,)
                    toreturn.append(mypath)
#            #Testing the cache
#            if validcache and toreturn != cached:
#                print "We return %s but %s was cached" %(str(toreturn),str(cached))
            self.cache_paths[node_id] = toreturn
            return toreturn

    def print_tree(self, string=False):
        """ Representation of tree in FilteredTree
        
        @param string: if set, instead of printing, return string for printing.
        """

        stack = [(self.root_id, "")]

        output = "_"*30 + "\n" + "FilteredTree cache\n" + "_"*30 + "\n"

        while stack != []:
            node_id, prefix = stack.pop()

            output += prefix + str(node_id) + '\n'

            for child_id in reversed(self.nodes[node_id]['children']):
                stack.append((child_id, prefix+" "))

        output += "_"*30 + "\n"

        if string:
            return output
        else:
            print output

    def get_all_nodes(self):
        nodes = list(self.nodes.keys())
        nodes.remove(self.root_id)
        return nodes

    def get_n_nodes(self, withfilters=[]):
        """
        returns quantity of displayed nodes in this tree
        if the withfilters is set, returns the quantity of nodes
        that will be displayed if we apply those filters to the current
        tree. It means that the currently applied filters are also taken into
        account.
        """
        return len(self.get_nodes(withfilters=withfilters))

    def get_nodes(self, withfilters=[]):
        """
        returns quantity of displayed nodes in this tree
        if the withfilters is set, returns the quantity of nodes
        that will be displayed if we apply those filters to the current
        tree. It means that the currently applied filters are also taken into
        account.
        """
        if withfilters == []:
            # Use current cache
            return self.get_all_nodes()
        #FIXME maybe allow caching multiple withfilters...
        elif len(withfilters) == 1 and withfilters[0] in self.filter_cache:
            return self.filter_cache[withfilters[0]]['nodes']
            
#        elif withfilters != []:
        else:
            # Filter on the current nodes

            filters = []
            for filter_name in withfilters:
                filt = self.fbank.get_filter(filter_name)
                if filt:
                    filters.append(filt)

            nodes = []
            for node_id in self.nodes:
                if node_id == self.root_id:
                    continue

                displayed = True
                for filt in filters:
                    displayed = filt.is_displayed(node_id)
                    if not displayed:
                        break
                
                if displayed:
                    nodes.append(node_id)

            return nodes

    def get_node_for_path(self, path):
        if not path or path == ():
            return None
        node_id = path[-1]
        #Both "if" should be benchmarked
        if path in self.get_paths_for_node(node_id):
#        if self.is_path_valid(path):
            return node_id
        else:
            return None

    def next_node(self, node_id, parent_id):
        if node_id == self.root_id:
            raise Exception("Calling next_node on the root node")

        if node_id not in self.nodes:
            raise Exception("Node %s is not displayed" % node_id)

        parents = self.nodes[node_id]['parents']
        if not parent_id:
            parent_id = parents[0]
        elif parent_id not in parents:
            raise Exception("Node %s does not have parent %s" % (node_id, parent_id))

        index = self.nodes[parent_id]['children'].index(node_id)
        if index+1 < len(self.nodes[parent_id]['children']):
            return self.nodes[parent_id]['children'][index+1]
        else:
            return None

    def node_all_children(self, node_id=None):
        if node_id is None:
            node_id = self.root_id
        return list(self.nodes[node_id]['children'])

    def node_has_child(self, node_id):
        return len(self.nodes[node_id]['children']) > 0

    def node_n_children(self, node_id, recursive=False):
        if node_id == None:
            node_id = self.root_id
        if not self.nodes.has_key(node_id):
            return 0
        if recursive:
            total = 0
            #We avoid recursion in a loop
            #because the dict might be updated in the meantime
            cids = list(self.nodes[node_id]['children'])   
            for cid in cids: 
                total += self.node_n_children(cid,recursive=True)
                total += 1 #we count the node itself ofcourse
            return total  
        else:
            return len(self.nodes[node_id]['children'])

    def node_nth_child(self, node_id, n):
        return self.nodes[node_id]['children'][n]

    def node_parents(self, node_id):
        if node_id not in self.nodes:
            raise IndexError('Node %s is not displayed' % node_id)
        parents = list(self.nodes[node_id]['parents'])
        if self.root_id in parents:
            parents.remove(self.root_id)
        return parents

    def get_current_state(self):
        """ Allows to connect LibLarch widget on fly to FilteredTree
        
        Sends 'added' signal/callback for every nodes that is currently
        in FilteredTree. After that, FilteredTree and TreeModel are
        in the same state
        """
        for node_id in self.nodes[self.root_id]['children']:
            self.send_add_tree(node_id, self.root_id)

#### FILTERS ##################################################################
    def list_applied_filters(self):
        return list(self.applied_filters)

    def apply_filter(self, filter_name, parameters=None, \
                    reset=None, refresh=None):
        """ Apply a new filter to the tree.

        @param filter_name: The name of an registrered filter from filters_bank
        @param parameters: Optional parameters to pass to the filter
        @param reset: Should other filters be removed?
        @param refresh: Should be refereshed the whole tree?
                (performance optimization)
        """
        if reset:
            self.applied_filters = []

        if parameters:
            filt = self.fbank.get_filter(filter_name)
            if filt:
                filt.set_parameters(parameters)
            else:
                raise ValueError("No filter of name %s in the bank" % filter_name)

        if filter_name not in self.applied_filters:
            self.applied_filters.append(filter_name)
            if refresh:
                self.refilter()
            toreturn = True
        else:
            toreturn = False

        return toreturn

    def unapply_filter(self, filter_name, refresh=True):
        """ Removes a filter from the tree.

        @param filter_name: The name of an already added filter to remove
        @param refresh: Should be refereshed the whole tree?
                (performance optimization)
        """
        if filter_name in self.applied_filters:
            self.applied_filters.remove(filter_name)
            if refresh:
                self.refilter()
            return True
        else:
            return False

    def reset_filters(self, refresh=True):
        """
        Clears all filters currently set on the tree.  Can't be called on 
        the main tree.
        """
        self.applied_filters = []
        if refresh:
            self.refilter()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import processqueue
from liblarch.treenode import _Node

class MainTree:
    """ Tree which stores and handle all requests """

    def __init__(self):
        """ Initialize MainTree.

        @param root - the "root" node which contains all nodes
        """

        self.nodes = {}
        self.pending_relationships = []

        self.__cllbcks = {}

        self.root_id = 'root'
        self.root = _Node(self.root_id)
        _Node._set_tree(self.root,self)

        self._queue = processqueue.SyncQueue()

    def __str__(self):
        return "<Tree: root = '%s'>" % self.root

    def get_root(self):
        """ Return root node """
        return self.root

####### INTERFACE FOR CALLBACKS ###############################################
    def register_callback(self, event, func):
        """ Store function and return unique key which can be used to
        unregister the callback later """

        if not self.__cllbcks.has_key(event):
            self.__cllbcks[event] = {}

        callbacks = self.__cllbcks[event]
        key = 0
        while callbacks.has_key(key):
            key += 1

        callbacks[key] = func
        return key

    def deregister_callback(self, event, key):
        """ Remove the callback identifed by key (from register_cllbck) """
        try:
            del self.__cllbcks[event][key]
        except KeyError:
            pass

    def _callback(self, event, node_id):
        """ Inform others about the event """
        #We copy the dict to not loop on it while it could be modified
        dic = dict(self.__cllbcks.get(event, {}))
        for func in dic.itervalues():
            func(node_id)

####### INTERFACE FOR HANDLING REQUESTS #######################################
    def add_node(self, node, parent_id=None, priority="low"):
        self._queue.push(self._add_node, node, parent_id, priority=priority)

    def remove_node(self, node_id, recursive=False):
        self._queue.push(self._remove_node, node_id, recursive)

    def modify_node(self, node_id, priority="low"):
        self._queue.push(self._modify_node, node_id, priority=priority)

    def new_relationship(self, parent_id, child_id):
        self._queue.push(self._new_relationship, parent_id, child_id)

    def break_relationship(self, parent_id, child_id):
        self._queue.push(self._break_relationship, parent_id, child_id)

    def refresh_all(self):
        """ Refresh all nodes """
        for node_id in self.nodes.keys():
            self.modify_node(node_id)

####### IMPLEMENTATION OF HANDLING REQUESTS ###################################
    def _create_relationship(self, parent_id, child_id):
        """ Create relationship without any checks """
        parent = self.nodes[parent_id]
        child = self.nodes[child_id]

        if child_id not in parent.children:
            parent.children.append(child_id)

        if parent_id not in child.parents:
            child.parents.append(parent_id)

        if child_id in self.root.children:
            self.root.children.remove(child_id)

    def _destroy_relationship(self, parent_id, child_id):
        """ Destroy relationship without any checks """
        parent = self.nodes[parent_id]
        child = self.nodes[child_id]

        if child_id in parent.children:
            parent.children.remove(child_id)

        if parent_id in child.parents:
            child.parents.remove(parent_id)

    def _is_circular_relation(self, parent_id, child_id):
        """ Would the new relation be circular?
        
        Go over every possible ancestors. If one of them is child_id,
        this would be circular relation.
        """

        visited = []
        ancestors = [parent_id]
        while ancestors != []:
            node_id = ancestors.pop(0)
            if node_id == child_id:
                return True
            
            if node_id not in self.nodes:
                continue
    
            for ancestor_id in self.nodes[node_id].parents:
                if ancestor_id not in visited:
                    ancestors.append(ancestor_id)

        return False
        
    def _add_node(self, node, parent_id):
        """ Add a node to the tree

        @param node - node to be added
        @param parent_id - parent to add or it will be add to root
        """
        node_id = node.get_id()
        if node_id in self.nodes:
            print "Error: Node '%s' already exists" % node_id
            return False

        _Node._set_tree(node,self)
        for relationship in node.pending_relationships:
            if relationship not in self.pending_relationships:
                self.pending_relationships.append(relationship)
        node.pending_relationships = []

        self.nodes[node_id] = node

        add_to_root = True
        parents_to_refresh = []
        children_to_refresh = []

        # Build pending relationships
        for rel_parent_id, rel_child_id in list(self.pending_relationships):
            # Adding as a child
            if rel_child_id == node_id and rel_parent_id in self.nodes:
                if not self._is_circular_relation(rel_parent_id, node_id):
                    self._create_relationship(rel_parent_id, node_id)
                    add_to_root = False
                    parents_to_refresh.append(rel_parent_id)
                else:
                    print "Error: Detected pending circular relationship", \
                        rel_parent_id, rel_child_id
                self.pending_relationships.remove((rel_parent_id, rel_child_id))

            # Adding as a parent
            if rel_parent_id == node_id and rel_child_id in self.nodes:
                if not self._is_circular_relation(node_id, rel_child_id):
                    self._create_relationship(node_id, rel_child_id)
                    children_to_refresh.append(rel_child_id)
                else:
                    print "Error: Detected pending circular relationship", \
                        rel_parent_id, rel_child_id
                self.pending_relationships.remove((rel_parent_id, rel_child_id))
        
        # Build relationship with given parent
        if parent_id is not None:
            if self._is_circular_relation(parent_id, node_id):
                raise Exception('Creating circular relationship between %s and %s' % \
                     (parent_id, node_id))
            if parent_id in self.nodes:
                self._create_relationship(parent_id, node_id)
                add_to_root = False
                parents_to_refresh.append(parent_id)
            else:
                self.pending_relationships.append((parent_id, node_id))

        # Add at least to root
        if add_to_root:
            self.root.children.append(node_id)

        # Send callbacks
        #updating the parent and the children is handled by the FT
        self._callback("node-added", node_id)
        
#        #The following callback is only needed in case we have a
#        #Flat filter applied.
#        for parent_id in parents_to_refresh:
#            self._callback("node-modified", parent_id)

        #this callback is really fast. No problem
#        for child_id in children_to_refresh:
#            #FIXME: why parent_id? this should be a bug!
#            #removing this doesn't affect the tests. Why is it useful?
#            self._callback("node-modified", child_id)

    def _remove_node(self, node_id, recursive=False):
        """ Remove node from tree """

        if node_id not in self.nodes:
            print "*** Warning *** Trying to remove a non-existing node"
            return

        # Do not remove root node
        if node_id is None:
            return

        # Remove pending relationships with this node
        for relation in list(self.pending_relationships):
            if node_id in relation:
                self.pending_relationships.remove(relation)

        node = self.nodes[node_id]

        # Handle parents
        for parent_id in node.parents:
            self._destroy_relationship(parent_id, node_id)
            self._callback('node-modified', parent_id)

        # Handle children
        for child_id in list(node.children):
            if recursive:
                self._remove_node(child_id, True)
            else:
                self._destroy_relationship(node_id, child_id)
                self._callback('node-modified', child_id)
                if self.nodes[child_id].parents == []:
                    self.root.children.append(child_id)

        if node_id in self.root.children:
            self.root.children.remove(node_id)

        self.nodes.pop(node_id)
        self._callback('node-deleted', node_id)

    def _modify_node(self, node_id):
        """ Force update of a node """
        if node_id != self.root_id and node_id in self.nodes:
            self._callback('node-modified', node_id)

    def _new_relationship(self, parent_id, child_id):
        """ Creates a new relationship 
        
        This method is used mainly from TreeNode"""

        if (parent_id, child_id) in self.pending_relationships:
            self.pending_relationships.remove((parent_id, child_id))

        if not parent_id or not child_id or parent_id == child_id:
            return False

        if parent_id not in self.nodes or child_id not in self.nodes:
            self.pending_relationships.append((parent_id, child_id))
            return True

        if self._is_circular_relation(parent_id, child_id):
            self._destroy_relationship(parent_id, child_id)
            raise Exception('Cannot build circular relationship between %s and %s' % (parent_id, child_id))


        self._create_relationship(parent_id, child_id)

        # Remove from root when having a new relationship
        if child_id in self.root.children:
            self.root.children.remove(child_id)

        self._callback('node-modified', parent_id)
        self._callback('node-modified', child_id)

    def _break_relationship(self, parent_id, child_id):
        """ Remove a relationship

        This method is used mainly from TreeNode """
        for rel_parent, rel_child in list(self.pending_relationships):
            if rel_parent == parent_id and rel_child == child_id:
                self.pending_relationships.remove((rel_parent, rel_child))

        if not parent_id or not child_id or parent_id == child_id:
            return False

        if parent_id not in self.nodes or child_id not in self.nodes:
            return False

        self._destroy_relationship(parent_id, child_id)

        # Move to root if beak the last parent
        if self.nodes[child_id].get_parents() == []:
            self.root.add_child(child_id)

        self._callback('node-modified', parent_id)
        self._callback('node-modified', child_id)


####### INTERFACE FOR READING STATE OF TREE ###################################
    def has_node(self, node_id):
        """ Is this node_id in this tree? """
        return node_id in self.nodes

    def get_node(self, node_id=None):
        """ Return node of tree or root node of this tree """
        if node_id in self.nodes:
            return self.nodes[node_id]
        elif node_id == self.root_id or node_id is None:
            return self.root
        else:
            raise ValueError("Node %s is not in the tree" % node_id)

    def get_node_for_path(self, path):
        """ Convert path into node_id
        
        @return node_id if path is valid, None otherwise
        """
        if not path or path == ():
            return None
        node_id = path[-1]
        if path in self.get_paths_for_node(node_id):
            return node_id
        else:
            return None
        return node_id

    def get_paths_for_node(self, node_id):
        """ Get all paths for node_id """
        if not node_id or node_id == self.root_id:
            return [()]
        elif node_id in self.nodes:
            node = self.nodes[node_id]
            if node.has_parent():
                paths = []
                for parent_id in node.get_parents():
                    if parent_id not in self.nodes:
                        continue
                    for path in self.get_paths_for_node(parent_id):
                        paths.append(path + (node_id,))
                return paths
            else:
                return [(node_id,)]
        else:
            raise ValueError("Cannot get path for non existing node %s" % node_id)

    def get_all_nodes(self):
        """ Return list of all nodes in this tree """
        return self.nodes.keys()

    def next_node(self, node_id, parent_id=None):
        """ Return the next sibling node or None if there is none
        
        @param  node_id - we look for siblings of this node
        @param parent_id - specify which siblings should be used, 
            if task has more parents. If None, random parent will be used
        """
        if node_id is None:
            raise ValueError('node_id should be different than None')

        node = self.get_node(node_id)
        parents_id = node.get_parents()
        if len(parents_id) == 0:
            parid = self.root_id
        elif parent_id in parents_id:
            parid = parent_id
        else:
            parid = parents_id[0]

        parent = self.get_node(parid)
        if not parent:
            raise ValueError('Parent does not exist')

        index = parent.get_child_index(node_id)
        if index == None:
            error = 'children are : %s\n' %parent.get_children()
            error += 'node %s is not a child of %s' %(node_id,parid)
            raise IndexError(error)

        if parent.get_n_children() > index+1:
            return parent.get_nth_child(index+1)
        else:
            return None

    def print_tree(self, string=False):
        output = self.root_id + "\n"
        stack = [(" ", child_id) for child_id in reversed(self.root.children)]

        while stack != []:
            prefix, node_id = stack.pop()
            output += prefix + node_id + "\n"
            prefix += " "
            for child_id in reversed(self.nodes[node_id].get_children()):
                stack.append((prefix, child_id))

        if string:
            return output
        else:
            print output,



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ---------------------------------------------------------------------------

class ViewCount:
    def __init__(self,tree,fbank, name = None, refresh = True):
        self.initialized = False
        self.ncount = {}
        self.tree = tree
        self.tree.register_callback("node-added", self.__modify)
        self.tree.register_callback("node-modified", self.__modify)
        self.tree.register_callback("node-deleted", self.__delete)
        
        self.fbank = fbank
        self.name = name
        
        self.applied_filters = []
        self.nodes = []
        self.cllbcks = []
        
        if refresh:
            self.__refresh()
                
    def __refresh(self):
        for node in self.tree.get_all_nodes():
            self.__modify(node)
        self.initialized = True
    
    def apply_filter(self,filter_name,refresh=True):
        if self.fbank.has_filter(filter_name):
            if filter_name not in self.applied_filters:
                self.applied_filters.append(filter_name)
                if refresh:
                    #If we are not initialized, we need to refresh with all existing nodes
                    if self.initialized:
                        for n in list(self.nodes):
                            self.__modify(n)
                    else:
                        self.__refresh()
        else:
            #FIXME: raise proper error
            print "There's no filter called %s" %filter_name
            
    def unapply_filter(self,filter_name):
        if filter_name in self.applied_filters:
            self.applied_filters.remove(filter_name)
            for node in self.tree.get_all_nodes():
                self.__modify(node)
    
    #there's only one callback: "modified"
    def register_cllbck(self, func):
        if func not in self.cllbcks:
            self.cllbcks.append(func)
            
    def unregister_cllbck(self,func):
        if func in self.cllbacks:
            self.cllbacks.remove(func)
    
    def get_n_nodes(self):
        return len(self.nodes)
        
    #Allow external update of a given node
    def modify(self,nid):
        self.__modify(nid)
        
    def __modify(self,nid):
        displayed = True
        for filtname in self.applied_filters:
            filt = self.fbank.get_filter(filtname)
            displayed &= filt.is_displayed(nid)
        if displayed:
            self.__add(nid)
        else:
            self.__delete(nid)
        

    
    def __delete(self,nid):
        if nid in self.nodes:
#            print "node %s removed from viewcount %s"%(nid,self.name)
            self.nodes.remove(nid)
            self.__callback()
            
    def __add(self,nid):
        if nid not in self.nodes:
#            print "node %s added to viewcount %s"%(nid,self.name)
            self.nodes.append(nid)
            self.__callback()
        
        
    def __callback(self):
        for c in self.cllbcks:
            c()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import gtk

class TreeModel(gtk.TreeStore):
    """ Local copy of showed tree """

    def __init__(self, tree, types):
        """ Initializes parent and create list of columns. The first colum
        is node_id of node """
        
        self.count = 0
        self.count2 = 0

        self.types = [[str, lambda node: node.get_id()]] + types
        only_types = [python_type for python_type, access_method in self.types]

        gtk.TreeStore.__init__(self, *only_types)
        self.cache_paths = {}
        self.cache_position = {}
        self.tree = tree

    def set_column_function(self, column_num, column_func):
        """ Replace function for generating certain column.

        Original use case was changing method of generating background
        color during runtime - background by tags or due date """

        if column_num < len(self.types):
            self.types[column_num][1] = column_func
            return True
        else:
            return False

    def connect_model(self):
        """ Register "signals", callbacks from liblarch.
        
        Also asks for the current status by providing add_task callback.
        We are able to connect to liblarch tree on the fly. """

        self.tree.register_cllbck('node-added-inview',self.add_task)
        self.tree.register_cllbck('node-deleted-inview',self.remove_task)
        self.tree.register_cllbck('node-modified-inview',self.update_task)
        self.tree.register_cllbck('node-children-reordered',self.reorder_nodes)

        # Request the current state
        self.tree.get_current_state()

    def my_get_iter(self, path):
        """ Because we sort the TreeStore, paths in the treestore are
        not the same as paths in the FilteredTree. We do the  conversion here.
        We receive a Liblarch path as argument and return a gtk.TreeIter"""
        #The function is recursive. We take iter for path (A,B,C) in cache.
        #If there is not, we take iter for path (A,B) and try to find C.
        if path == ():
            return None
        nid = str(path[-1])
        self.count += 1
        #We try to use the cache
        iter = self.cache_paths.get(path,None)
        toreturn = None
        if iter and self.iter_is_valid(iter) and nid == self.get_value(iter,0):
            self.count2 += 1
            toreturn = iter
        else:
            root = self.my_get_iter(path[:-1])
            #This is a small ad-hoc optimisation.
            #Instead of going through all the children nodes
            #We go directly at the last known position.
            pos = self.cache_position.get(path,None)
            if pos:
                iter = self.iter_nth_child(root,pos)
                if iter and self.get_value(iter,0) == nid:
                    toreturn = iter
            if not toreturn:
                if root:
                    iter = self.iter_children(root)
                else:
                    iter = self.get_iter_first()
                while iter and self.get_value(iter,0) != nid:
                    iter = self.iter_next(iter)
            self.cache_paths[path] = iter
            toreturn = iter
#        print "%s / %s" %(self.count2,self.count)
#        print "my_get_iter %s : %s" %(nid,self.get_string_from_iter(toreturn))
        return toreturn

    def print_tree(self):
        """ Print TreeStore as Tree into console """

        def push_to_stack(stack, level, iterator):
            """ Macro which adds a new element into stack if it is possible """
            if iterator is not None:
                stack.append((level, iterator))

        stack = []
        push_to_stack(stack, 0, self.get_iter_first())

        print "+"*50
        print "Treemodel print_tree: "
        while stack != []:
            level, iterator = stack.pop()

            print "=>"*level, self.get_value(iterator, 0)

            push_to_stack(stack, level, self.iter_next(iterator))
            push_to_stack(stack, level+1, self.iter_children(iterator))
        print "+"*50

### INTERFACE TO LIBLARCH #####################################################

    def add_task(self, node_id, path):
        """ Add new instance of node_id to position described at path.

        @param node_id: identification of task
        @param path: identification of position
        """
        node = self.tree.get_node(node_id)

        # Build a new row
        row = []
        for python_type, access_method in self.types:
            value = access_method(node)
            row.append(value)

        # Find position to add task
        iter_path = path[:-1]

        iterator = self.my_get_iter(iter_path)
        self.cache_position[path] = self.iter_n_children(iterator)
        it = self.insert(iterator, -1, row)
        
        # Show the new task if possible
#        self.row_has_child_toggled(self.get_path(it), it)

    def remove_task(self, node_id, path):
        """ Remove instance of node.

        @param node_id: identification of task
        @param path: identification of position
        """
        it = self.my_get_iter(path)
        if not it:
            raise Exception("Trying to remove node %s with no iterator"%node_id)
        actual_node_id = self.get_value(it, 0)
        assert actual_node_id == node_id
        self.remove(it)
        self.cache_position.pop(path)

    def update_task(self, node_id, path):
        """ Update instance of node by rebuilding the row.

        @param node_id: identification of task
        @param path: identification of position
        """
        #We cannot assume that the node is in the tree because
        #update is asynchronus
        #Also, we should consider that missing an update is not critical
        #and ignoring the case where there is no iterator
        if self.tree.is_displayed(node_id):
            node = self.tree.get_node(node_id)
            #That call to my_get_iter is really slow!
            iterator = self.my_get_iter(path)
        
            if iterator:
                for column_num, (python_type, access_method) in enumerate(self.types):
                    value = access_method(node)
                    self.set_value(iterator, column_num, value)

    def reorder_nodes(self, node_id, path, neworder):
        """ Reorder nodes.
        
        This is deprecated signal. In the past it was useful for reordering
        showed nodes of tree. It was possible to delete just the last
        element and therefore every element must be moved to the last position
        and then deleted.

        @param node_id: identification of root node
        @param path: identification of poistion of root node
        @param neworder: new order of children of root node
        """

        if path is not None:
            it = self.my_get_iter(path)
        else:
            it = None
        self.reorder(it, neworder)
        self.print_tree()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""
filters_bank stores all of GTG's filters in centralized place
"""


class Filter:
    def __init__(self, func, req):
        self.func = func
        self.dic = {}
        self.tree = req

    def set_parameters(self, dic):
        if dic:
            self.dic = dic
    
    def is_displayed(self, node_id):
        if self.tree.has_node(node_id):
            task = self.tree.get_node(node_id)
        else:
            return False

        if self.dic:
            value = self.func(task, parameters=self.dic)
        else:
            value = self.func(task)

        if 'negate' in self.dic and self.dic['negate']:
            value = not value

        return value
        
    def get_parameters(self, param):
        return self.dic.get(param, None)

    def is_flat(self):
        """ Should be the final list flat """
        return self.get_parameters('flat')
    
class FiltersBank:
    """
    Stores filter objects in a centralized place.
    """

    def __init__(self,tree):
        """
        Create several stock filters:

        workview - Tasks that are active, workable, and started
        active - Tasks of status Active
        closed - Tasks of status closed or dismissed
        notag - Tasks with no tags
        """
        self.tree = tree
        self.available_filters = {}
        self.custom_filters = {}

    ##########################################
        
    def get_filter(self,filter_name):
        """ Get the filter object for a given name """
        if self.available_filters.has_key(filter_name):
            return self.available_filters[filter_name]
        elif self.custom_filters.has_key(filter_name):
            return self.custom_filters[filter_name]
        else:
            return None
            
    def has_filter(self,filter_name):
        return self.available_filters.has_key(filter_name) \
            or self.custom_filters.has_key(filter_name)
    
    def list_filters(self):
        """ List, by name, all available filters """
        liste = self.available_filters.keys()
        liste += self.custom_filters.keys()
        return liste
    
    def add_filter(self,filter_name,filter_func,parameters=None):
        """
        Adds a filter to the filter bank 
        Return True if the filter was added
        Return False if the filter_name was already in the bank
        """
        if filter_name not in self.list_filters():
            negate = False
            if filter_name.startswith('!'):
                negate = True
                filter_name = filter_name[1:]
            else:
                filter_obj = Filter(filter_func,self.tree)
                filter_obj.set_parameters(parameters)
            self.custom_filters[filter_name] = filter_obj
            return True
        else:
            return False
        
    def remove_filter(self,filter_name):
        """
        Remove a filter from the bank.
        Only custom filters that were added here can be removed
        Return False if the filter was not removed
        """
        if not self.available_filters.has_key(filter_name):
            if self.custom_filters.has_key(filter_name):
                self.custom_filters.pop(filter_name)
                return True
            else:
                return False
        else:
            return False



#!/usr/bin/python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Liblarch - a library to handle directed acyclic graphs
# Copyright (c) 2011-2012 - Lionel Dricot & Izidor Matušov
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from distutils.core import setup

setup(
    version = '2.1.0',
    url = 'https://live.gnome.org/liblarch',
    author = 'Lionel Dricot & Izidor Matušov',
    author_email = 'gtg-contributors@lists.launchpad.net',
    license = 'LGPLv3',
    name = 'liblarch',
    packages = ['liblarch', 'liblarch_gtk'],
    description = 'Liblarch is a python library built to easily handle '\
	'data structures such as lists, trees and directed acyclic graphs '\
	'and represent them as GTK TreeWidget or in other forms.',
)
